@@ -3,14 +3,14 @@
  *
  * Apply a WIM image directly to an NTFS volume using libntfs-3g.  Restore as
  * much information as possible, including security data, file attributes, DOS
- * names, and alternate data streams.
+ * names, alternate data streams, and object IDs.
  *
- * Note: because NTFS-3g offers inode-based interfaces, we actually don't need
+ * Note: because NTFS-3G offers inode-based interfaces, we actually don't need
  * to deal with paths at all!  (Other than for error messages.)
  */
 
 /*
- * Copyright (C) 2012, 2013, 2014, 2015 Eric Biggers
+ * Copyright (C) 2012-2016 Eric Biggers
  *
  * This file is free software; you can redistribute it and/or modify it under
  * the terms of the GNU Lesser General Public License as published by the Free
@@ -35,6 +35,7 @@
 #include <string.h>
 
 #include <ntfs-3g/attrib.h>
+#include <ntfs-3g/object_id.h>
 #include <ntfs-3g/reparse.h>
 #include <ntfs-3g/security.h>
 
@@ -46,6 +47,7 @@
 #include "wimlib/error.h"
 #include "wimlib/metadata.h"
 #include "wimlib/ntfs_3g.h"
+#include "wimlib/object_id.h"
 #include "wimlib/reparse.h"
 #include "wimlib/security.h"
 #include "wimlib/security_descriptor.h"
@@ -65,6 +67,7 @@ ntfs_3g_get_supported_features(const char *target,
 	supported_features->reparse_points            = 1;
 	supported_features->security_descriptors      = 1;
 	supported_features->short_names               = 1;
+	supported_features->object_ids                = 1;
 	supported_features->timestamps                = 1;
 	supported_features->case_sensitive_filenames  = 1;
 	return 0;
@@ -117,6 +120,8 @@ sid_size(const wimlib_SID *sid)
  *   bug.  However, removing the DACL is not a valid workaround because this
  *   changes the meaning of the security descriptor--- an empty DACL allows no
  *   access, whereas a "null" DACL allows all access.
+ * - Versions before 2016.2.22 reject security descriptors containing SIDs with
+ *   too many subauthorities.  We do not work around this.
  *
  * If the security descriptor was fixed, this function returns an allocated
  * buffer containing the fixed security descriptor, and its size is updated.
@@ -299,7 +304,7 @@ ntfs_3g_restore_dos_name(ntfs_inode *ni, ntfs_inode *dir_ni,
 		goto out_close;
 	}
 
-	/* Unlike most other NTFS-3g functions, ntfs_set_ntfs_dos_name()
+	/* Unlike most other NTFS-3G functions, ntfs_set_ntfs_dos_name()
 	 * changes the directory's last modification timestamp...
 	 * Change it back.  */
 	return ntfs_3g_restore_timestamps(vol, dentry->d_parent->d_inode);
@@ -332,8 +337,8 @@ ntfs_3g_restore_reparse_point(ntfs_inode *ni, const struct wim_inode *inode,
 				"tag.  The preceding error may have been caused "
 				"by a known bug in libntfs-3g where it does not "
 				"correctly validate non-Microsoft reparse "
-				"points.  This bug may be fixed in the 2016 "
-				"release of libntfs-3g.");
+				"points.  This bug was fixed in NTFS-3G version "
+				"2016.2.22.");
 		}
 		return WIMLIB_ERR_SET_REPARSE_DATA;
 	}
@@ -412,6 +417,25 @@ ntfs_3g_set_metadata(ntfs_inode *ni, const struct wim_inode *inode,
 	sd = wim_get_current_security_data(ctx->common.wim);
 	one_dentry = inode_first_extraction_dentry(inode);
 
+	/* Object ID */
+	{
+		u32 len;
+		const void *object_id = inode_get_object_id(inode, &len);
+		if (unlikely(object_id != NULL) &&
+		    ntfs_set_ntfs_object_id(ni, object_id, len, 0))
+		{
+			if (errno == EEXIST) {
+				WARNING("Duplicate object ID on file \"%s\"",
+					dentry_full_path(one_dentry));
+			} else {
+				ERROR_WITH_ERRNO("Failed to set object ID on "
+						 "\"%s\" in NTFS volume",
+						 dentry_full_path(one_dentry));
+				return WIMLIB_ERR_NTFS_3G;
+			}
+		}
+	}
+
 	/* Attributes  */
 	if (!(extract_flags & WIMLIB_EXTRACT_FLAG_NO_ATTRIBUTES)) {
 		u32 attrib = inode->i_attributes;
@@ -455,7 +479,8 @@ ntfs_3g_set_metadata(ntfs_inode *ni, const struct wim_inode *inode,
 					"\nThis error occurred because libntfs-3g thinks "
 					"the security descriptor is invalid.  If you "
 					"are extracting a Windows 10 image, this may be "
-					"caused by a known bug in libntfs-3g.  See: "
+					"caused by a known bug in libntfs-3g.  This bug "
+					"was fixed in NTFS-3G version 2016.2.22.  See: "
 					"https://wimlib.net/forums/viewtopic.php?f=1&t=4 "
 					"for more information.\n\n");
 			}
@@ -552,9 +577,10 @@ ntfs_3g_create_directories(struct wim_dentry *root,
 
 	/* Set the DOS name of any directory that has one.  In addition, create
 	 * empty attributes for directories that have them.  Note that creating
-	 * an empty reparse point attribute must happen *after* setting the
-	 * DOS name in order to work around a case where
-	 * ntfs_set_ntfs_dos_name() fails with EOPNOTSUPP.  */
+	 * an empty reparse point attribute must happen *after* setting the DOS
+	 * name in order to work around a case where ntfs_set_ntfs_dos_name()
+	 * fails with EOPNOTSUPP.  This bug was fixed in NTFS-3G version
+	 * 2016.2.22.  */
 	list_for_each_entry(dentry, dentry_list, d_extraction_list_node) {
 		const struct wim_inode *inode = dentry->d_inode;
 
@@ -993,7 +1019,7 @@ ntfs_3g_extract(struct list_head *dentry_list, struct apply_ctx *_ctx)
 	struct wim_dentry *root;
 	int ret;
 
-	/* For NTFS-3g extraction mode we require that the dentries to extract
+	/* For NTFS-3G extraction mode we require that the dentries to extract
 	 * form a single tree.  */
 	root = list_first_entry(dentry_list, struct wim_dentry,
 				d_extraction_list_node);
@@ -1001,7 +1027,7 @@ ntfs_3g_extract(struct list_head *dentry_list, struct apply_ctx *_ctx)
 	/* Mount the NTFS volume.  */
 	vol = ntfs_mount(ctx->common.target, 0);
 	if (!vol) {
-		ERROR_WITH_ERRNO("Failed to mount \"%s\" with NTFS-3g",
+		ERROR_WITH_ERRNO("Failed to mount \"%s\" with NTFS-3G",
 				 ctx->common.target);
 		return WIMLIB_ERR_NTFS_3G;
 	}
@@ -1042,7 +1068,7 @@ ntfs_3g_extract(struct list_head *dentry_list, struct apply_ctx *_ctx)
 
 out_unmount:
 	if (ntfs_umount(ctx->vol, FALSE) && !ret) {
-		ERROR_WITH_ERRNO("Failed to unmount \"%s\" with NTFS-3g",
+		ERROR_WITH_ERRNO("Failed to unmount \"%s\" with NTFS-3G",
 				 ctx->common.target);
 		ret = WIMLIB_ERR_NTFS_3G;
 	}
@@ -1050,7 +1076,7 @@ out_unmount:
 }
 
 const struct apply_operations ntfs_3g_apply_ops = {
-	.name			= "NTFS-3g",
+	.name			= "NTFS-3G",
 	.get_supported_features = ntfs_3g_get_supported_features,
 	.extract                = ntfs_3g_extract,
 	.context_size           = sizeof(struct ntfs_3g_apply_ctx),
