@@ -6,7 +6,7 @@
  */
 
 /*
- * Copyright (C) 2012, 2013, 2014, 2015 Eric Biggers
+ * Copyright (C) 2012-2016 Eric Biggers
  *
  * This file is free software; you can redistribute it and/or modify it under
  * the terms of the GNU Lesser General Public License as published by the Free
@@ -31,6 +31,7 @@
 #include <errno.h>
 
 #include <ntfs-3g/attrib.h>
+#include <ntfs-3g/object_id.h>
 #include <ntfs-3g/reparse.h>
 #include <ntfs-3g/security.h>
 #include <ntfs-3g/volume.h>
@@ -38,20 +39,21 @@
 #include "wimlib/alloca.h"
 #include "wimlib/assert.h"
 #include "wimlib/blob_table.h"
-#include "wimlib/capture.h"
 #include "wimlib/dentry.h"
 #include "wimlib/encoding.h"
 #include "wimlib/endianness.h"
 #include "wimlib/error.h"
 #include "wimlib/ntfs_3g.h"
+#include "wimlib/object_id.h"
 #include "wimlib/paths.h"
 #include "wimlib/reparse.h"
+#include "wimlib/scan.h"
 #include "wimlib/security.h"
 
-/* NTFS-3g 2013 renamed MS_RDONLY to NTFS_MNT_RDONLY.  We can't check for the
+/* NTFS-3G 2013 renamed MS_RDONLY to NTFS_MNT_RDONLY.  We can't check for the
  * existence of NTFS_MNT_RDONLY at compilation time because it's an enum.  We
  * also can't check for MS_RDONLY being missing because it's also a system
- * constant.  So check if the NTFS-3g specific MS_IGNORE_HIBERFILE is defined;
+ * constant.  So check if the NTFS-3G specific MS_IGNORE_HIBERFILE is defined;
  * if yes, then we need to use the old MS_RDONLY.  */
 #ifdef MS_IGNORE_HIBERFILE
 #  define NTFS_MNT_RDONLY MS_RDONLY
@@ -269,7 +271,7 @@ set_attr_sort_key(ntfs_inode *ni, struct ntfs_location *loc)
 /*
  * Add a new stream to the specified inode, with duplicate checking.
  *
- * This works around a problem where NTFS-3g can list multiple unnamed data
+ * This works around a problem where NTFS-3G can list multiple unnamed data
  * streams for a single file.  In this case we can only keep one.  We'll prefer
  * one that is nonempty.
  */
@@ -436,6 +438,22 @@ out_put_actx:
 	return ret;
 }
 
+static noinline_for_stack int
+load_object_id(ntfs_inode *ni, struct wim_inode *inode)
+{
+	OBJECT_ID_ATTR attr;
+	int len;
+
+	len = ntfs_get_ntfs_object_id(ni, (char *)&attr, sizeof(attr));
+	if (likely(len == -ENODATA || len == 0))
+		return 0;
+	if (len < 0)
+		return WIMLIB_ERR_NTFS_3G;
+	if (!inode_set_object_id(inode, &attr, len))
+		return WIMLIB_ERR_NOMEM;
+	return 0;
+}
+
 /* Load the security descriptor of an NTFS inode into the corresponding WIM
  * inode and the WIM image's security descriptor set.  */
 static noinline_for_stack int
@@ -593,7 +611,7 @@ struct readdir_ctx {
 	size_t path_len;
 	struct dos_name_map dos_name_map;
 	struct ntfs_volume_wrapper *volume;
-	struct capture_params *params;
+	struct scan_params *params;
 	int ret;
 };
 
@@ -604,7 +622,7 @@ ntfs_3g_build_dentry_tree_recursive(struct wim_dentry **root_p,
 				    size_t path_len,
 				    int name_type,
 				    struct ntfs_volume_wrapper *volume,
-				    struct capture_params *params);
+				    struct scan_params *params);
 
 static int
 filldir(void *_ctx, const ntfschar *name, const int name_nchars,
@@ -658,7 +676,7 @@ static int
 ntfs_3g_recurse_directory(ntfs_inode *ni, char *path, size_t path_len,
 			  struct wim_dentry *parent,
 			  struct ntfs_volume_wrapper *volume,
-			  struct capture_params *params)
+			  struct scan_params *params)
 {
 	int ret;
 	s64 pos = 0;
@@ -703,7 +721,7 @@ ntfs_3g_build_dentry_tree_recursive(struct wim_dentry **root_ret,
 				    size_t path_len,
 				    int name_type,
 				    struct ntfs_volume_wrapper *volume,
-				    struct capture_params *params)
+				    struct scan_params *params)
 {
 	u32 attributes;
 	int ret;
@@ -735,13 +753,13 @@ ntfs_3g_build_dentry_tree_recursive(struct wim_dentry **root_ret,
 	if (unlikely(attributes & FILE_ATTRIBUTE_ENCRYPTED)) {
 		if (params->add_flags & WIMLIB_ADD_FLAG_NO_UNSUPPORTED_EXCLUDE)
 		{
-			ERROR("Can't archive \"%s\" because NTFS-3g capture mode "
+			ERROR("Can't archive \"%s\" because NTFS-3G capture mode "
 			      "does not support encrypted files and directories", path);
 			ret = WIMLIB_ERR_UNSUPPORTED_FILE;
 			goto out;
 		}
 		params->progress.scan.cur_path = path;
-		ret = do_capture_progress(params, WIMLIB_SCAN_DENTRY_UNSUPPORTED, NULL);
+		ret = do_scan_progress(params, WIMLIB_SCAN_DENTRY_UNSUPPORTED, NULL);
 		goto out;
 	}
 
@@ -776,6 +794,13 @@ ntfs_3g_build_dentry_tree_recursive(struct wim_dentry **root_ret,
 			goto out;
 	}
 
+	/* Load the object ID.  */
+	ret = load_object_id(ni, inode);
+	if (ret) {
+		ERROR_WITH_ERRNO("Error reading object ID of \"%s\"", path);
+		goto out;
+	}
+
 	/* Scan the data streams.
 	 *
 	 * Note: directories should not have an unnamed data stream, but they
@@ -787,7 +812,7 @@ ntfs_3g_build_dentry_tree_recursive(struct wim_dentry **root_ret,
 	if (ret)
 		goto out;
 
-	/* Reparse-point fixups are a no-op because in NTFS-3g capture mode we
+	/* Reparse-point fixups are a no-op because in NTFS-3G capture mode we
 	 * only allow capturing an entire volume. */
 	if (params->add_flags & WIMLIB_ADD_FLAG_RPFIX &&
 	    inode_is_symlink(inode))
@@ -813,16 +838,16 @@ ntfs_3g_build_dentry_tree_recursive(struct wim_dentry **root_ret,
 out_progress:
 	params->progress.scan.cur_path = path;
 	if (root == NULL)
-		ret = do_capture_progress(params, WIMLIB_SCAN_DENTRY_EXCLUDED, NULL);
+		ret = do_scan_progress(params, WIMLIB_SCAN_DENTRY_EXCLUDED, NULL);
 	else
-		ret = do_capture_progress(params, WIMLIB_SCAN_DENTRY_OK, inode);
+		ret = do_scan_progress(params, WIMLIB_SCAN_DENTRY_OK, inode);
 out:
 	if (ni)
 		ntfs_inode_close(ni);
 	if (unlikely(ret)) {
 		free_dentry_tree(root, params->blob_table);
 		root = NULL;
-		ret = report_capture_error(params, ret, path);
+		ret = report_scan_error(params, ret, path);
 	}
 	*root_ret = root;
 	return ret;
@@ -830,8 +855,7 @@ out:
 
 int
 ntfs_3g_build_dentry_tree(struct wim_dentry **root_ret,
-			  const char *device,
-			  struct capture_params *params)
+			  const char *device, struct scan_params *params)
 {
 	struct ntfs_volume_wrapper *volume;
 	ntfs_volume *vol;
